<!DOCTYPE html>
<html>
<head><title>Confirmation</title>
<link rel="stylesheet" href="{{url_for('login',asset='css')}}">
</head>
<body>
    <div id="login">
        <h1>Confirmation</h1>
        <form name="logout" method="POST">
            Please confirm that you would like to log out: <input type="submit" value="Confirm">
        </form>
    </div>
</body>
</html>
